/*
 * autopilot.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 * based on code of same name 	
 * Copyright (C) 2000 Leigh Morresi (leigh@signal-x.com)
 */


/*
 * Monifications - look at CTS line as this is asserted by the
 * serial driver throughout the duration (DSR if often 
 * disconnected to stop cradle drain). If CTS is deasserted 
 * it signifies either handshaking - or that the hot sync has  
 * failed. We terminate the child and go back to waiting if
 * timeout reads spaced delay apart return CTS deasserted 
 */

static char rcsid[] =
  "$Id: autopilot.c,v 1.3 2001/09/14 12:11:50 root Exp root $";


#define SHELL_NAME "sh"
#define SHELL_PATH "/bin/sh"

/*
 * $Log: autopilot.c,v $
 * Revision 1.3  2001/09/14 12:11:50  root
 * *** empty log message ***
 *
 * Revision 1.2  2001/09/14 10:26:01  root
 * *** empty log message ***
 *
 * Revision 1.1  2001/09/14 10:25:54  root
 * Initial revision
 *
 */

/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef PALMDEVICE
#define PALMDEVICE "/dev/pilot"
#endif

#include <syslog.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/wait.h>
#include <termios.h>

/* global */
char *device = PALMDEVICE;
char *command = NULL;
int delay = 2;
int timeout = 10;

/* print command usage and exit */
void
usage ()
{
  fprintf (stderr, "usage: autopilot [-n seconds] [-t timeout] [-p port] -c <command>\n");
  fprintf (stderr, "  -n interval between scan for pilot in seconds."
           " (default %d)\n", delay);
  fprintf (stderr, "  -t number of intervals to detect timeout."
           " (default %d)\n", timeout);
  fprintf (stderr, "  -p port to listen on."
           " (default %s)\n", PALMDEVICE);
  fprintf (stderr,
           "  -c command to run on active ie 'pilot-mail -p /dev/pilot'\n");
  exit (1);
}

/* handle command line options */
void
parse_args (int argc, char **argv)
{
  const char *flags = "n:c:p:t:h";
  int c;
  extern char *optarg;

  while ((c = getopt (argc, argv, flags)) != EOF)
    {
      switch (c)
        {
        case 'n':
          delay = atoi (optarg);
          break;
        case 't':
          timeout = atoi (optarg);
          break;
        case 'c':
          command = optarg;
          break;
        case 'p':
          device = optarg;
          break;
        case '?':
        case 'h':
          usage ();
          exit (1);
          break;
        }
    }

}


void
run_child (int fd)
{
  int pid;
  unsigned int arg;
  int deasserted = 0;
  int status;

  switch (pid = fork ())
    {
    case 0:
      {
        char *args[4];

	close(fd);

        args[0] = SHELL_NAME;
        args[1] = "-c";
        args[2] = command;
        args[3] = NULL;

        signal (SIGINT, SIG_DFL);
        signal (SIGQUIT, SIG_DFL);

        execv (SHELL_PATH, args);
        syslog (LOG_ERR, "run_child: execve %s -c %s failed %m", SHELL_PATH,
                command);
        exit (-1);
      }

      break;
    case -1:
      syslog (LOG_ERR, "run_child: fork failed %m");
      return;
    }


  syslog (LOG_ERR, "run_child: launched %d", pid);
  for (;;)
    {

      switch (waitpid (pid, &status, WNOHANG))
        {
        case 0:
          break;
        case -1:
          /* Possible race condition gets us here */
          syslog (LOG_ERR, "run_child: waitpid %m");
          return;
        default:
          syslog (LOG_ERR, "run_child: collected %d", pid);
          return;
        }




      if (ioctl (fd, TIOCMGET, &arg))
        syslog (LOG_ERR, "TIOCMGET failed %m");

      if (arg & TIOCM_CTS)
        {
          deasserted = 0;
        }
      else
        {
          deasserted++;
        }


      if (deasserted == timeout)
        {
          /*Kill child, waitpid will clean up and quit this loop */
          syslog (LOG_ERR, "Zapping child %d", pid);
          kill (pid,SIGTERM);
        }

      if (deasserted == (2 * timeout))
        {
          /*Kill child, waitpid will clean up and quit this loop */
          syslog (LOG_ERR, "Zapping child %d", pid);
          kill (pid,SIGKILL);
        }

/*
      if (deasserted == (3 * timeout))
        {
        char *args[4];
        syslog (LOG_ERR, "Self destruct initiated...", pid);
        args[0] = SHELL_NAME;
        args[1] = "-c";
        args[2] = "/software/scripts/sr";
        args[3] = NULL;

        signal (SIGINT, SIG_DFL);
        signal (SIGQUIT, SIG_DFL);

        execv (SHELL_PATH, args);
        syslog (LOG_ERR, "run_child: execve %s -c %s failed %m", SHELL_PATH,
                command);
        exit (-1);
        }
*/

      sleep (delay);

    }

}

void do_fd(int fd)
{
          struct termios tcn;
int i;

  tcgetattr(fd,&tcn);

        cfmakeraw(&tcn);

        tcn.c_oflag = 0;
        tcn.c_iflag = IGNBRK | IGNPAR;

        tcn.c_cflag = CREAD | CLOCAL | CS8;

        tcn.c_lflag = NOFLSH;

        (void) cfsetspeed(&tcn, B9600);

        for (i = 0; i < 16; i++)
                tcn.c_cc[i] = 0;

        tcsetattr(fd, TCSANOW, &tcn);
}


int
main (int argc, char *argv[])
{
  int fd;
  unsigned int arg;
  int state, old_state;
  int i;

  if (getenv ("PILOTPORT") != NULL)
    device = getenv ("PILOTPORT");

  parse_args (argc, argv);

  if (!command)
    {
      fprintf (stderr,
               "You must specify a command to run with the -c option.\n");
      usage ();
      exit (1);
    }

  if (!device)
    {
      fprintf (stderr, "No device set PILOTPORT or use -p option");
      usage ();
      exit (1);
    }

  fd = open (device, O_RDONLY | O_NDELAY | O_NOCTTY);

  if (fd < 0)
    {
      fprintf (stderr, "Can't open device %s: %s\n", device,
               strerror (errno));
      usage ();
      exit (1);
    }

  do_fd(fd);

        if ((i = fcntl(fd, F_GETFL, 0)) != -1) {
                i &= ~O_NONBLOCK;
                fcntl(fd, F_SETFL, i);
        }



#ifndef DEBUG
  switch (fork ())
    {
    case 0:
      break;
    case -1:
      perror ("fork failed");
      exit (1);
    default:
      exit (0);
    }

  daemon (0, 0);
#endif

  openlog ("autosync", LOG_CONS, LOG_DAEMON);

  signal (SIGCHLD, SIG_IGN);

  state = 0;

  for (;;)
    {
      old_state = state;

      if (ioctl (fd, TIOCMGET, &arg))
        syslog (LOG_ERR, "TIOCMGET failed %m");

      /*Raise RTS to let autosync.prc detect the cradle */
      arg |= TIOCM_RTS;

      if (ioctl (fd, TIOCMSET, &arg))
        syslog (LOG_ERR, "TIOCMSET failed %m");

      /* If CTS asserted run the child */
      state = (arg & TIOCM_CTS) ? 1 : 0;

      if ((state) && (state != old_state)) {
        run_child (fd);
	do_fd(fd);
      }

      sleep (delay);
    }

  return 0;
}
