/*
 * cleaner.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 *
 */

static char rcsid[] =
  "$Id: cleaner.c,v 1.3 2001/11/23 02:20:56 root Exp root $";

/*
 * $Log: cleaner.c,v $
 * Revision 1.3  2001/11/23 02:20:56  root
 * #
 *
 * Revision 1.2  2001/11/23 02:20:53  root
 * #
 *
 * Revision 1.1  2001/11/23 01:15:55  root
 * Initial revision
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/param.h>
#include <stdarg.h>
#include <strings.h>
#include <string.h>
#include <sys/stat.h>
#include <malloc.h>
#include <unistd.h>

#include "db.h"

#include "pi-source.h"
#include "pi-socket.h"
#include "pi-dlp.h"
#include "pi-file.h"
#include "pi-header.h"


typedef struct Palmdb_struct
{
  char name[34];
#ifdef QUICKSCANCODE
  char type[5];
  char creator[5];
#else
  unsigned long type;
  unsigned long creator;
#endif
  int serial;
}
 *Palmdb;
#define PALMDB_SIZE (sizeof(struct Palmdb_struct))

DB *files_db;
DB *notfound_db;
DB *ovwr_db;

static int install_serial = 0;
static int forced = 0;

#define DBERROR(d,o,r,fmt...) my_dberror(d,#d,#o,r,__FILE__,__LINE__,fmt)
#define CRASH_AND_BURN(fmt...) crash_and_burn(__FILE__,__LINE__,fmt)



void
my_dberror (DB * db, char *dbn, char *op, int ret, char *file, int line,
            char *fmt, ...)
{
  char buf[1025];
  int i;
  va_list ap;

  if (db)
    {
      i = sprintf (buf, "%s line %d: %s->%s:", file, line, dbn, op);
    }
  else
    {
      i = sprintf (buf, "%s line %d: %s:", file, line, op);
    }

  if (fmt)
    {
      va_start (ap, fmt);
      vsnprintf (buf + i, sizeof (buf) - 1 - i, fmt, ap);
      va_end (ap);
      buf[sizeof (buf) - 1] = 0;
    }


  if (db)
    {
      db->err (db, ret, buf);
    }
  else
    {
      fprintf (stderr, "%s: %s(%d)\n", buf, db_strerror (ret), ret);
    }
  exit (1);
}

void
crash_and_burn (char *file, int line, char *fmt, ...)
{
  char buf[1025];
  int i;
  va_list ap;

  i = sprintf (buf, "crash_and_burn: %s line %d: ", file, line);

  if (fmt)
    {
      va_start (ap, fmt);
      vsnprintf (buf + i, sizeof (buf) - 1 - i, fmt, ap);
      va_end (ap);
    }
  buf[sizeof (buf) - 1] = 0;
  fprintf (stderr, "%s\n", buf);

  exit (1);
}


Palmdb
file_to_palmdb (char *file)
{
  Palmdb ret;
#ifdef QUICKSCANCODE
  FILE *f = fopen (file, "r");
  char buf[68];
  if (!f)
    return NULL;

  if (fread (buf, 1, 68, f) != 68)
    {
      fclose (f);
      return NULL;
    }
  fclose (f);

  ret = malloc (PALMDB_SIZE);
  bzero (ret, PALMDB_SIZE);
  bcopy (buf, ret->name, 32);
  bcopy (buf + 60, ret->type, 4);
  bcopy (buf + 64, ret->creator, 4);
  ret->type[4] = 0;
  ret->creator[4] = 0;
#else
  struct pi_file *pf = pi_file_open (file);
  struct DBInfo info = { 0 };

  if (!pf)
    return NULL;

  if (pi_file_get_info (pf, &info) < 0)
    {
      pi_file_close (pf);
      return NULL;
    }
  pi_file_close (pf);

  ret = malloc (PALMDB_SIZE);
  bzero (ret, PALMDB_SIZE);


  strcpy (ret->name, info.name);

  ret->type = info.type;
  ret->creator = info.creator;

#endif
  ret->name[32] = 0;
  ret->serial = 0;

  return ret;
}

int
check_if_suitable (char *src, char *dst)
{
  struct pi_file *spf = pi_file_open (src);
  struct pi_file *dpf = pi_file_open (dst);
  struct DBInfo si = { 0 }, di =
  {
  0};
  int se = -2, de = -1;

  int i;
  unsigned long stype, dtype, duid, suid;
  int sid, did;
  void *sbuf, *dbuf;
  int ssize, dsize;
  int sattrs, scat;
  int dattrs, dcat;
  int sapp_info_size = -1, dapp_info_size = -2;
  void *sapp_info, *dapp_info;
#ifdef DEBUG
  printf ("Comparing %s and %s:\t", src, dst);
#endif
  if ((!spf) || (!dpf))
    goto err;
  if (pi_file_get_info (spf, &si) < 0)
    goto err;
  if (pi_file_get_info (spf, &di) < 0)
    goto err;

  si.backupDate = di.backupDate = 0;
  si.flags &= ~(dlpDBFlagCopyPrevention | dlpDBFlagOpen | dlpDBFlagBackup);
  di.flags &= ~(dlpDBFlagCopyPrevention | dlpDBFlagOpen | dlpDBFlagBackup);

#if 1
  si.modnum = di.modnum = 0;
  si.createDate = di.createDate = 0;
  si.modifyDate = di.modifyDate = 0;
#endif

  if (bcmp (&si, &di, sizeof (si)))
    goto err;

#ifdef DEBUG
  printf (" HM");
#endif


  if (pi_file_get_app_info (spf, &sapp_info, &sapp_info_size) < 0)
    sapp_info_size = 0;
  if (pi_file_get_app_info (dpf, &dapp_info, &dapp_info_size) < 0)
    dapp_info_size = 0;

  if (sapp_info_size)
    {
      if ((sapp_info_size != dapp_info_size)
          || bcmp (sapp_info, dapp_info, sapp_info_size)) goto err;
#ifdef DEBUG
      printf (" AI");
#endif

    }

  sapp_info_size = -1;
  dapp_info_size = -2;

  if (pi_file_get_sort_info (spf, &sapp_info, &sapp_info_size) < 0)
    sapp_info_size = 0;
  if (pi_file_get_sort_info (dpf, &dapp_info, &dapp_info_size) < 0)
    dapp_info_size = 0;

  if (sapp_info_size)
    {
      if ((sapp_info_size != dapp_info_size)
          || bcmp (sapp_info, dapp_info, sapp_info_size)) goto err;
#ifdef DEBUG
      printf (" SI");
#endif

    }



  pi_file_get_entries (spf, &se);
  pi_file_get_entries (dpf, &de);

  if (se != de)
    goto err;
#ifdef DEBUG
  printf (" NM");
#endif

  if (si.flags & dlpDBFlagResource)
    {
      for (i = 0; i < se; ++i)
        {
          /*buf lives inside the spf struct */
          if (pi_file_read_resource (spf, i, &sbuf, &ssize, &stype, &sid) < 0)
            goto err;
          if (pi_file_read_resource (dpf, i, &dbuf, &dsize, &dtype, &did) < 0)
            goto err;

          if ((ssize != dsize) || (stype != dtype) || (sid != did)
              || bcmp (sbuf, dbuf, ssize))
            {
#ifdef DEBUG
              printf (" !R %d", i);
#endif
              goto err;
            }
        }
    }
  else
    {
      for (i = 0; i < se; ++i)
        {
          /*buf lives inside the spf struct */
          if (pi_file_read_record
              (spf, i, &sbuf, &ssize, &sattrs, &scat, &suid) < 0)
            goto err;
          if (pi_file_read_record
              (dpf, i, &dbuf, &dsize, &dattrs, &dcat, &duid) < 0)
            goto err;

          if ((ssize != dsize) || (sattrs != dattrs) || (scat != dcat)
              || (suid != duid) || bcmp (sbuf, dbuf, ssize))
            {
#ifdef DEBUG
              printf (" !R %d", i);
#endif
              goto err;
            }
        }
    }

#ifdef DEBUG
  printf (" RM");
#endif



  if (spf)
    pi_file_close (spf);
  if (dpf)
    pi_file_close (dpf);
#ifdef DEBUG
  printf ("\n");
#endif
  return 1;


err:
#ifdef DEBUG
  printf ("! splat\n");
#endif
  if (spf)
    pi_file_close (spf);
  if (dpf)
    pi_file_close (dpf);
  return 0;
}




void
do_install (char *from)
{
  FILE *in, *out;
  char buf[1024];
  int red;

  in = fopen (from, "r");
  if (!in)
    CRASH_AND_BURN ("Couldn't open %s\n", from);
  sprintf (buf, "install/cleaner%06d.pdb", install_serial++);
  out = fopen (buf, "w");
  if (!out)
    CRASH_AND_BURN ("Couldn't open %s\n", buf);

#ifdef DEBUG
  printf ("%s->%s\n", from, buf);
#endif

  while ((red = fread (buf, 1, sizeof (buf), in)) > 0)
    {
      fwrite (buf, 1, red, out);
    }
  fclose (in);
  fclose (out);

}

void
find_copies (char *file)
{
  DBT key = { 0 };
  DBT data = { 0 };
  int ret;
  int found = 0;

  Palmdb d;


  d = file_to_palmdb (file);

  if (!d)
    return;

  key.data = d;
  key.size = PALMDB_SIZE;


  do
    {
      ret = files_db->get (files_db, NULL, &key, &data, 0);
      if (ret == DB_NOTFOUND)
        break;
      if (ret)
        DBERROR (files_db, get, ret, NULL);
      found = check_if_suitable (data.data, file);
      if (found)
        break;
      d->serial++;
    }
  while (!ret);

  if (!found)
    {

      key.data = file;
      key.size = strlen (file) + 1;
      ret = files_db->get (ovwr_db, NULL, &key, &data, 0);
      if (!ret)
        {
          forced++;
          do_install (data.data);
        }
      else if (ret == DB_NOTFOUND)
        {

          key.data = d;
          key.size = PALMDB_SIZE;
          d->serial = 0;

          do
            {
              ret = files_db->get (files_db, NULL, &key, &data, 0);
              if (ret == DB_NOTFOUND)
                break;
              if (ret)
                DBERROR (files_db, get, ret, NULL);

              key.data = file;
              key.size = strlen (file) + 1;
              found++;

              if (
                  (ret =
                   notfound_db->put (notfound_db, NULL, &key, &data, 0)))
                DBERROR (notfound_db, put, ret, NULL);

              d->serial++;
            }
          while (!ret);

          if (!found)
            {
              key.data = file;
              key.size = strlen (file) + 1;
              data.data = "";
              data.size = 1;
              if (
                  (ret =
                   notfound_db->put (notfound_db, NULL, &key, &data, 0)))
                DBERROR (notfound_db, put, ret, NULL);
            }



        }
      else
        {
          DBERROR (ovwr_db, get, ret, NULL);
        }

    }
  else
    {
      do_install (data.data);
    }

  free (d);
}

void
add_file_to_db (char *file)
{
  DBT key = { 0 };
  DBT data = { 0 };
  int ret;

  Palmdb d = file_to_palmdb (file);


  if (!d)
    return;

  key.data = d;
  key.size = PALMDB_SIZE;

  while ((ret = files_db->get (files_db, NULL, &key, &data, 0)) == 0)
    d->serial++;

  switch (ret)
    {
    case DB_NOTFOUND:
      break;
    default:
      DBERROR (files_db, get, ret, NULL);
    }

  data.data = file;
  data.size = strlen (file) + 1;

  if ((ret = files_db->put (files_db, NULL, &key, &data, 0)))
    DBERROR (files_db, put, ret, NULL);

  free (d);
}

void
do_dir (char *dir, void (*func) (char *))
{
  DIR *d = opendir (dir);
  char buf[MAXPATHLEN];
  struct dirent *e;
  struct stat s;
  char *ptr;
  int i;

  if (!d)
    return;


  while ((e = readdir (d)))
    {
      if (!strcmp (e->d_name, "."))
        continue;
      if (!strcmp (e->d_name, ".."))
        continue;

      strcpy (buf, dir);
      strcat (buf, "/");
      strcat (buf, e->d_name);

      if (stat (buf, &s))
        continue;

      if (S_ISDIR (s.st_mode))
        do_dir (buf, func);

      i = strlen (ptr = e->d_name);
      ptr += i - 4;
      if (!strcmp (ptr, ".prc"))
        func (buf);
      if (!strcmp (ptr, ".pdb"))
        func (buf);
      if (!strcmp (ptr, ".PDB"))
        func (buf);
      if (!strcmp (ptr, ".PRC"))
        func (buf);

    }

}
void
read_ovwr (void)
{
  char buf[2048], *ptr;
  char *src, *dst;
  DBT key = { 0 }, data =
  {
  0};
  int ret;
  FILE *f;

  f = fopen ("cleaner.overwrite", "r");
  if (!f)
    return;

  while (!feof (f))
    {
      fgets (buf, sizeof (buf) - 1, f);
      buf[sizeof (buf) - 1] = 0;

      while ((ptr = index (buf, '\n')))
        *ptr = 0;


      do
        {
          dst = index (buf, '"');
          if (!dst)
            break;
          dst++;
          ptr = index (dst, '"');
          if (!ptr)
            break;
          *ptr = 0;
          ptr++;
          src = index (ptr, '"');
          if (!src)
            break;
          src++;
          ptr = index (src, '"');
          if (!ptr)
            break;
          *ptr = 0;


#ifdef DEBUG
          printf ("Auto overwriting %s from %s\n", dst, src);
#endif

          key.data = dst;
          key.size = strlen (dst) + 1;
          data.data = src;
          data.size = strlen (src) + 1;

          if ((ret = ovwr_db->put (ovwr_db, NULL, &key, &data, 0)))
            DBERROR (ovwr_db, put, ret, NULL);
        }
      while (0);
    }

  fclose (f);

}



void
walk_lost (void)
{
  DBC *c;
  DBT key = { 0 }, data =
  {
  0};
  int ret;
  FILE *f;

  f = fopen ("cleaner.notfound", "w");

  if (!f)
    return;

  if (ret = notfound_db->cursor (notfound_db, NULL, &c, 0))
    {
      DBERROR (notfound_db, cursor, ret, NULL);
      return;
    }


  while (1)
    {
      ret = c->c_get (c, &key, &data, DB_NEXT);
      if (ret == DB_NOTFOUND)
        break;
      if (ret)
        DBERROR (notfound_db, c_get, ret, NULL);
      fprintf (f, "\"%s\" \"%s\"\n", key.data, data.data);
    }

  c->c_close (c);

  fclose (f);


}

void
usage (void)
{
  CRASH_AND_BURN (NULL);
}

int
main (int argc, char *argv[])
{
  int ret;
  int i;

  if ((ret = db_create (&files_db, NULL, 0)))
    {
      DBERROR (NULL, db_create, ret, NULL);
      CRASH_AND_BURN (NULL);
    }
  if (
      (ret =
       files_db->open (files_db, NULL, NULL, DB_BTREE, DB_CREATE, 0600)))
    {
      DBERROR (files_db, open, ret, NULL);
      files_db->close (files_db, 0);
      CRASH_AND_BURN (NULL);
    }


  if ((ret = db_create (&notfound_db, NULL, 0)))
    {
      DBERROR (NULL, db_create, ret, NULL);
      CRASH_AND_BURN (NULL);
    }

  if ((ret = notfound_db->set_flags (notfound_db, DB_DUP)))
    {
      DBERROR (notfound_db, set_flags, ret, "DB_DUP");
      notfound_db->close (notfound_db, 0);
      CRASH_AND_BURN (NULL);
    }


  if (
      (ret =
       notfound_db->open (notfound_db, NULL, NULL, DB_BTREE, DB_CREATE,
                          0600)))
    {
      DBERROR (notfound_db, open, ret, NULL);
      notfound_db->close (notfound_db, 0);
      CRASH_AND_BURN (NULL);
    }


  if ((ret = db_create (&ovwr_db, NULL, 0)))
    {
      DBERROR (NULL, db_create, ret, NULL);
      CRASH_AND_BURN (NULL);
    }
  if (
      (ret =
       notfound_db->open (ovwr_db, NULL, NULL, DB_BTREE, DB_CREATE, 0600)))
    {
      DBERROR (ovwr_db, open, ret, NULL);
      notfound_db->close (ovwr_db, 0);
      CRASH_AND_BURN (NULL);
    }


  read_ovwr ();

  if (argc > 1)
    {
      chdir (getenv ("HOME"));
      chdir ("pilot");
      for (i = 1; i < argc; ++i)
        {
          do_dir (argv[i], add_file_to_db);
        }
      do_dir ("hotsync", find_copies);
    }
  else
    {
      usage ();
    }

  walk_lost ();
  files_db->close (files_db, 0);
  notfound_db->close (notfound_db, 0);

  fprintf (stderr, "%d databases refreshed (%d forced)\n", install_serial,
           forced);

  return 0;
}
