/*
 * emaildb.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 *
 */

static char rcsid[] = "$Id: emaildb.c,v 1.2 2001/09/08 16:59:57 root Exp $";

/*
 * $Log: emaildb.c,v $
 * Revision 1.2  2001/09/08 16:59:57  root
 * #
 *
 * Revision 1.1  2001/09/08 16:57:22  root
 * #
 *
 */

#include <stdio.h>
#include <strings.h>
#include <malloc.h>
#include <string.h>

#include <db.h>

#define EMDB_REMOVE 1
#define EMDB_INPUT 2
static DB *db = NULL;

struct AddressAppInfo;
extern add_email_cb (char *, char *, int, int, struct AddressAppInfo *aai);

int
my_isspace (char c)
{
  switch (c)
    {
    case '\t':
    case '\n':
    case '"':
    case ' ':
    case ',':
      return 1;
    }
  return 0;
}

static void
db_open (void)
{
  int ret;

  if (db)
    return;

  if ((ret = db_create (&db, NULL, 0)))
    {
      abort ();
    }


  if ((ret = db->open (db, NULL, NULL, DB_BTREE, DB_CREATE, 0600)))
    {
      db->close (db, 0);
      abort ();
    }


}

db_worker (int op, char *email, char *name)
{
  DBT key = { 0 };
  DBT data = { 0 };

  if (!strlen (email))
    return;

  key.data = email;
  key.size = strlen (email) + 1;

  if (name)
    {
      if (!strlen (name))
        return;
      data.data = name;
      data.size = strlen (name) + 1;
    }
  else
    {
      if (op == EMDB_INPUT)
        return;
      data.data = "";
      data.size = 0;
    }


  switch (op)
    {

    case EMDB_INPUT:
/*
        printf ("+A%sA B%sB\n", data.data, key.data);
*/
      db->put (db, NULL, &key, &data, 0);
      break;
    case EMDB_REMOVE:
/*
        printf ("-B%sB\n", key.data);
*/
      db->del (db, NULL, &key, 0);
      break;
    }

}


static void
worker (int op, char *s)
{
  char *p1, *p2, *p3;
  if (!s)
    return;


  if ((!(p1 = index (s, '<'))) || (!(p2 = index (s, '>'))))
    {
      /*Then we must assume whole thing is email addr */
      while (my_isspace (*s))
        s++;


      p1 = s + strlen (s);
      p1--;

      while (my_isspace (*p1) && (p1 > s))
        *(p1--) = 0;


      db_worker (op, s, NULL);

      return;

    }



  while (my_isspace (*s))
    s++;

  p3 = p1 + 1;
  *p1 = ' ';

  while (my_isspace (*p1) && (p1 > s))
    *(p1--) = 0;
  if (p1 == s)
    return;
  *p2 = 0;

  db_worker (op, p3, s);
}

static void
parser (int op, char *cs)
{
  char *t, *u;
  int s = 0, e = 0, l;
  if (!cs)
    return;

  db_open ();

  t = strdup (cs);

  while (1)
    {

      switch (t[e])
        {
        case 0:
        case '>':
        case ',':
        case '\n':
        case '\t':
          l = 1 + e - s;
          u = malloc (l + 1);
          memcpy (u, t + s, l);
          u[l] = 0;

          worker (op, u);

          free (u);
          s = e + 1;
          break;
        }

      if (!t[e])
        break;
      e++;
    }

  free (t);
}

void
emdb_input_free (char *s)
{
  parser (EMDB_INPUT, s);
  free (s);
}


void
emdb_input (char *s)
{
  parser (EMDB_INPUT, s);
}

void
emdb_remove_free (char *s)
{
  parser (EMDB_REMOVE, s);
  free (s);
}

void
emdb_remove (char *s)
{
  parser (EMDB_REMOVE, s);
}

void
emdb_cleanup (void)
{
  db->close (db, 0);
  db = NULL;
}

void
emdb_dump (int sd, int idb, struct AddressAppInfo *aai)
{
  DBC *c;
  DBT key = { 0 };
  DBT data = { 0 };
  int ret;

  db_open ();

  if (ret = db->cursor (db, NULL, &c, 0))
    abort ();

  while (1)
    {
      char *ckey, *cdata;
      ret = c->c_get (c, &key, &data, DB_NEXT);
      if (ret == DB_NOTFOUND)
        break;
      if (ret)
        abort ();


      ckey = strdup (key.data);
      cdata = strdup (data.data);

/*
        printf ("+A%sA B%sB\n", data.data, key.data);
*/

      add_email_cb (cdata, ckey, sd, idb, aai);

      free (cdata);
      free (ckey);
    }

}
