/*
 * lib.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 *
 */

static char rcsid[] = "$Id: lib.c,v 1.3 2001/09/14 10:24:39 root Exp root $";

/*
 * $Log: lib.c,v $
 * Revision 1.3  2001/09/14 10:24:39  root
 * #
 *
 * Revision 1.2  2001/09/08 16:59:57  root
 * #
 *
 * Revision 1.1  2001/09/08 16:57:22  root
 * #
 *
 */

#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <malloc.h>
#include <dirent.h>

#include <time.h>


#include <pi-socket.h>
#include <pi-dlp.h>

extern int address_main (int sock, int argc, char *argv[]);
extern int xfer_main (int sock, int argc, char *argv[]);
extern int syncal_main (int sock, int argc, char *argv[]);
extern int syncmail_main (int sock, int argc, char *argv[]);

#ifdef PIC
void
plugin_version (int *major_version, int *minor_version)
{
  *major_version = 0;
  *minor_version = 99;
}

int
plugin_get_name (char *name, int len)
{
  strncpy (name, "UberSync 0.99", len);
  return 0;
}

int
plugin_get_help_name (char *name, int len)
{
  strncpy (name, "About UberSync", len);
  return 0;
}

int
plugin_help (char **text, int *width, int *height)
{
  /* We could also pass back *text=NULL */
  *text = strdup (
           /*-------------------------------------------*/
                   "UberSync plugin for J-Pilot was written by\n"
                   "James McKenzie (c) 2001.\n" "james@fishsoup.dhs.org\n");
  *height = 200;
  *width = 300;
  return 0;
}

#endif

int
pilot_open (char *device)
{
  struct pi_sockaddr addr = { 0 };
  int ret, pilot_socket;

  if (0 == (pilot_socket = pi_socket (PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP)))
    {
      perror ("pi_socket");
      exit (1);
    }

  addr.pi_family = PI_AF_SLP;
  strcpy (addr.pi_device, device);

  ret = pi_bind (pilot_socket, (struct sockaddr *) &addr, sizeof (addr));
  if (ret == -1)
    {
      fprintf (stderr, "Unable to bind to port '%s'.\n", device);
      exit (1);
    }

  fprintf (stderr,
           "Waiting for connection on %s (press the HotSync button now)...\n",
           device);

  ret = pi_listen (pilot_socket, 1);
  if (ret == -1)
    {
      perror ("pi_listen");
      exit (1);
    }

  pilot_socket = pi_accept (pilot_socket, 0, 0);
  if (pilot_socket == -1)
    {
      perror ("pi_accept");
      exit (1);
    }

  fprintf (stderr, "Connected.\n");

  return pilot_socket;
}

void
pilot_close (int pilot_socket)
{
  if (pilot_socket == 0)
    return;

  dlp_EndOfSync (pilot_socket, 0);
  pi_close (pilot_socket);
  pilot_socket = 0;
}


void
print_argv (char **argv)
{

  while (*argv)
    {
      printf ("A%sA ", *argv);
      argv++;
    }
  printf ("\n");
}



char **
argv_push (char **argv, char *s)
{
  int *ip = (int *) &argv[-1];
  int i = *ip;


  argv[i] = malloc (1024);
  strcpy (argv[i], s);
  i++;

  argv = realloc (argv - 1, (i + 2) * sizeof (char *));
  argv++;

  ip = (int *) &argv[-1];
  argv[i] = NULL;

  *ip = i;

  return argv;
}

int
argc_of_argv (char **argv)
{
  int i = 0;
  while (*(argv++))
    i++;
  return i;
}

char **
new_argv (void)
{
  char **argv = malloc (sizeof (char *) * 2);
  int *ip;

  extern int optind;

  optind = 0;

  argv++;
  ip = (int *) &argv[-1];
  *ip = 0;
  argv[0] = NULL;

  return argv;


}

void
free_argv (char **argv)
{
  int *ip = (int *) &argv[-1];
  int i;

  for (i = 0; i < *ip; ++i)
    {
      free (argv[i]);
    }

  free (argv - 1);
}

char **
argv_ptr (char **argv)
{
  char **ret;
  int *ip = (int *) &argv[-1];

  ret = malloc (sizeof (char *) * ((*ip) + 1));

  memcpy (ret, argv, sizeof (char *) * ((*ip) + 1));

  return ret;
}

#ifdef PIC

int
plugin_sync (int sock)
{

#else
int
main (int argc, char *argv[])
{
  int sock;
#endif
  char **a, **ptr;

#ifndef PIC
  sock = pilot_open ("/dev/pilot");
  if (sock < 0)
    return -1;
#endif

  printf ("Doing mail\n");
  {
    a = new_argv ();
    a = argv_push (a, "pilot-mailsync");

    ptr = argv_ptr (a);
    syncmail_main (sock, argc_of_argv (ptr), ptr);
    free (ptr);

    free_argv (a);
  }

  printf ("Address book ->\n");
  {
    a = new_argv ();

    a = argv_push (a, "pilot-addresses");
    a = argv_push (a, "-w");
    a = argv_push (a, "/dev/null");


    ptr = argv_ptr (a);
    address_main (sock, argc_of_argv (ptr), ptr);
    free (ptr);

    free_argv (a);

  }

  printf ("Address book <-\n");
  {
    a = new_argv ();

    a = argv_push (a, "pilot-addresses");
    a = argv_push (a, "-c");
    a = argv_push (a, "E-mail");
    a = argv_push (a, "-r");
    a = argv_push (a, "/dev/null");


    ptr = argv_ptr (a);
    address_main (sock, argc_of_argv (ptr), ptr);
    free (ptr);

    free_argv (a);

  }


  printf ("Installing\n");


  {
    DIR *d;
    struct dirent *e;
    int i = 0;
    char buf[1024];
    char dir[1024];

    strcpy (dir, getenv ("HOME"));
    strcat (dir, "/pilot/install");
    d = opendir (dir);

    if (d)
      {
        a = new_argv ();

        a = argv_push (a, "pilot-xfer");
        a = argv_push (a, "--install");

        while (e = readdir (d))
          {
            if (strcmp (e->d_name, ".") && strcmp (e->d_name, ".."))
              {
                strcpy (buf, dir);
                strcat (buf, "/");
                strcat (buf, e->d_name);
                i++;
                a = argv_push (a, buf);
              }
          }

        if (i)
          {
            ptr = argv_ptr (a);
            xfer_main (sock, argc_of_argv (ptr), ptr);
            free (ptr);
          }

        rewinddir (d);
        while (e = readdir (d))
          {
            if (strcmp (e->d_name, ".") && strcmp (e->d_name, ".."))
              {
                strcpy (buf, dir);
                strcat (buf, "/");
                strcat (buf, e->d_name);
                unlink (buf);
              }
          }


        closedir (d);
        free_argv (a);
      }
  }

  printf ("Backing up\n");
  {
    char buf[1024];

    a = new_argv ();
    a = argv_push (a, "pilot-xfer");
    a = argv_push (a, "--exclude");
    strcpy (buf, getenv ("HOME"));
    strcat (buf, "/pilot/exclude");
    a = argv_push (a, buf);
    a = argv_push (a, "--sync");
    strcpy (buf, getenv ("HOME"));
    strcat (buf, "/pilot/hotsync");
    a = argv_push (a, buf);

    ptr = argv_ptr (a);
    xfer_main (sock, argc_of_argv (ptr), ptr);
    free (ptr);

    free_argv (a);
  }


  printf ("Doing calendar\n");

  {
    a = new_argv ();
    a = argv_push (a, "syncal");
    a = argv_push (a, "-v");

    ptr = argv_ptr (a);
    syncal_main (sock, argc_of_argv (ptr), ptr);
    free (ptr);

    free_argv (a);
  }
  emdb_cleanup ();

#ifndef PIC
  pilot_close (sock);
#endif
  return 0;
}
