/*
 * pilot-addresses.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 *
 */

static char rcsid[] =
  "$Id: pilot-addresses.c,v 1.2 2001/09/08 16:59:57 root Exp $";

/*
 * $Log: pilot-addresses.c,v $
 * Revision 1.2  2001/09/08 16:59:57  root
 * #
 *
 * Revision 1.1  2001/09/08 16:57:22  root
 * #
 *
 */

/*
 * pilot-addresses.c:  Palm address transfer utility
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "pi-source.h"
#include "pi-socket.h"
#include "pi-dlp.h"
#include "pi-address.h"
#include "pi-header.h"

#define PILOTPORT "/dev/pilot"

/* Define prototypes */
static int inchar (FILE * in);
static int read_field (char *dest, FILE * in);
static void outchar (char c, FILE * out);
static int write_field (FILE * out, char *source, int more);
static int match_category (char *buf, struct AddressAppInfo *aai);
static int match_phone (char *buf, struct AddressAppInfo *aai);
static int read_file (FILE * in, int sd, int db, struct AddressAppInfo *aai);
static int write_file (FILE * out, int sd, int db,
                       struct AddressAppInfo *aai);
static void Help (char *progname);
static void Version (void);

/* Yet more hair: reorganize fields to match visible appearence */
static int realentry[19] =
  { 0, 1, 13, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 14, 15, 16, 17, 18 };

static char *tableheads[22] =
  { "Last name", "First name", "Company", "Work", "Home",
  "Fax", "Other", "Email", "Address", "City", "State",
  "Zip Code", "Country", "Title", "Custom1", "Custom2", "Custom3",
  "Custom4", "Note", "Main", "Pager", "Mobile"
};

static int tableformat = 0;
static int tabledelim = 1;
static char tabledelims[5] = { '\n', ',', ';', '\t' };
static int encodechars = 0;

extern void emdb_remove (char *s);
extern void emdb_remove_free (char *s);


static int
inchar (FILE * in)
{
  int c;

  c = getc (in);
  if (encodechars && c == '\\')
    {
      c = getc (in);
      switch (c)
        {
        case 'b':
          c = '\b';
          break;
        case 'f':
          c = '\f';
          break;
        case 'n':
          c = '\n';
          break;
        case 't':
          c = '\t';
          break;
        case 'r':
          c = '\r';
          break;
        case 'v':
          c = '\v';
          break;
        case '\\':
          c = '\\';
          break;
        default:
          ungetc (c, in);
          c = '\\';
          break;
        }
    }
  return c;
}


static int
read_field (char *dest, FILE * in)
{
  int c;

  do
    {                           /* Absorb whitespace */
      c = getc (in);
    }
  while ((c != EOF)
         && ((c == ' ') || (c == '\t') || (c == '\n') || (c == '\r')));

  if (c == '"')
    {
      c = inchar (in);

      while (c != EOF)
        {
          if (c == ',' || c == '\n')
            {
//   if (c == '"') {
              c = inchar (in);
              if (c != '"')
                break;
            }
          *dest++ = c;
          c = inchar (in);
        }
    }
  else
    {
      while (c != EOF)
        {
          if (c == ',' || (tableformat && c == tabledelims[tabledelim]))
            {
              break;
            }
          *dest++ = c;
          c = inchar (in);
        }
    }
  *dest++ = '\0';

  while ((c != EOF) && ((c == ' ') || (c == '\t'))) /* Absorb whitespace */
    c = getc (in);

  if (c == ',')
    return 1;
  else if (c == ';')
    return 2;
  else if (c == '\t')
    return 3;
  else if (c == EOF)
    return -1;                  /* No more */
  else
    return 0;
}

static void
outchar (char c, FILE * out)
{

  if (encodechars)
    {
      switch (c)
        {
        case '"':
          putc ('"', out);
          putc ('"', out);
          break;
        case '\b':
          putc ('\\', out);
          putc ('b', out);
          break;
        case '\f':
          putc ('\\', out);
          putc ('f', out);
          break;
        case '\n':
          putc ('\\', out);
          putc ('n', out);
          break;
        case '\t':
          putc ('\\', out);
          putc ('t', out);
          break;
        case '\r':
          putc ('\\', out);
          putc ('r', out);
          break;
        case '\v':
          putc ('\\', out);
          putc ('v', out);
          break;
        case '\\':
          putc ('\\', out);
          putc ('\\', out);
          break;
        default:
          putc (c, out);
          break;
        }
    }
  else
    {
      putc (c, out);
      if (c == '"')
        putc ('"', out);
    }
}

static int
write_field (FILE * out, char *source, int more)
{

  putc ('"', out);
  while (*source)
    {
      outchar (*source, out);
      source++;
    }
  putc ('"', out);

  putc (tabledelims[more], out);
#if 0
  if (more == 1)
    putc (',', out);
  else if (more == 2)
    putc (';', out);
  else if (more == 3)
    putc ('\t', out);
  else if (more == 0)
    putc ('\n', out);
#endif
  return 0;
}

static int
match_category (char *buf, struct AddressAppInfo *aai)
{
  int i;

  for (i = 0; i < 16; i++)
    if (strcasecmp (buf, aai->category.name[i]) == 0)
      return i;
  return atoi (buf);            /* 0 is default */
}

static int
match_phone (char *buf, struct AddressAppInfo *aai)
{
  int i;

  for (i = 0; i < 8; i++)
    {
      if (strcasecmp (buf, aai->phoneLabels[i]) == 0)
        return i;
    }
  return atoi (buf);            /* 0 is default */
}

static int defaultcategory = 0;

extern emdb_dump (int sd, int idb, struct AddressAppInfo *aai);
int
add_email_cb (char *name, char *email, int sd, int db,
              struct AddressAppInfo *aai)
{
  struct Address a;
  int i, l;
  char buf[0xffff];
  int category, attribute;
  char *fn, *ln, *em;

  memset (&a, 0, sizeof (a));
  a.showPhone = 0;
  attribute = 0;

  fprintf(stderr,"Adding record: name=%s email=%s\n",name,email);
  ln = rindex (name, ' ');

  if (!ln)
    {
      ln = strdup (name);
      fn = strdup ("");
    }
  else
    {
      *ln = 0;
      ln = strdup (ln + 1);
      fn = strdup (name);
    }

  a.entry[0] = ln;
  a.entry[1] = fn;

/*
  a.phoneLabel[0]= 3;
  a.phoneLabel[1]= 5;
  a.phoneLabel[2]= 4;
  a.phoneLabel[3]= 1;
*/
  a.phoneLabel[4] = match_phone ("E-mail", aai);
  a.entry[7] = em = strdup (email);


  l = pack_Address (&a, (unsigned char *) buf, sizeof (buf));
#ifdef DEBUG
  dumpdata (buf, l);
#endif
  if (ln)
    free (ln);
  if (fn)
    free (fn);
  if (em)
    free (em);
  dlp_WriteRecord (sd, db, attribute, 0, defaultcategory,
                   (unsigned char *) buf, l, 0);




  return 0;
}

static int augment = 0;
static int tablehead = 0;

static int
write_file (FILE * out, int sd, int db, struct AddressAppInfo *aai)
{
  struct Address a;
  int i, j, l;
  char buf[0xffff];
  int category, attribute;

#if 0
  if (tablehead)
    {
      write_field (out, "Category", tabledelim);
      for (j = 0; j < 19; j++)
        {
          write_field (out, tableheads[realentry[j]], tabledelim);
        }
      write_field (out, "Private-Flag", 0);
    }
#endif


  for (i = 0;
       (j =
        dlp_ReadRecordByIndex (sd, db, i, (unsigned char *) buf, 0,
                               &l, &attribute, &category)) >= 0; i++)
    {


      if (attribute & dlpRecAttrDeleted)
        continue;
      unpack_Address (&a, (unsigned char *) buf, l);
      dlp_AddSyncLogEntry (sd, ".");

      for (j = 0; j < 19; j++)
        {
          if (a.entry[j])
            emdb_remove_free (a.entry[j]);
        }

    }

  return 0;
}

static char *progname;

static void
Help (char *progname)
{
  PalmHeader (progname);

  fprintf (stderr,
           "   Usage: %s [-aeqDT] [-t delim] [-p port] [-c category]\n",
           progname);
  fprintf (stderr, "             [-d category] -r|-w [<file>]\n\n");
  fprintf (stderr,
           "     -t delim    = include category, use delimiter (3=tab, 2=;, 1=,)\n");
  fprintf (stderr, "     -T          = write header with titles\n");
  fprintf (stderr,
           "     -q          = do not prompt for HotSync button press\n");
  fprintf (stderr,
           "     -a          = augment records with additional information\n");
  fprintf (stderr,
           "     -e          = escape special characters with backslash\n");
  fprintf (stderr,
           "     -p port     = use device file <port> to communicate with Palm\n");
  fprintf (stderr,
           "     -c category = install to category <category> by default\n");
  fprintf (stderr,
           "     -d category = delete old Palm records in <category>\n");
  fprintf (stderr,
           "     -D          = delete all Palm records in all categories\n");
  fprintf (stderr,
           "     -r file     = read records from <file> and install them to Palm\n");
  fprintf (stderr,
           "     -w file     = get records from Palm and write them to <file>\n\n");
  exit (2);
}

static void
Version (void)
{
  PalmHeader (progname);
  fprintf (stderr,
           "   Type 'man 7 %s' at your shell for more information.\n\n",
           progname);
  exit (0);
}

int
address_main (int sd, int argc, char *argv[])
{
  struct pi_sockaddr addr;
  int db;
  int l;
  struct PilotUser U;
  int ret;
  char buf[0xffff];
  struct AddressAppInfo aai;
  char *defaultcategoryname = 0;
  char *deletecategory = 0;
  char *progname = argv[0];

/*   char *device = argv[1]; */
  int deleteallcategories = 0;
  int quiet = 0;
  int mode = 0;
  int c;
  extern char *optarg;
  extern int optind;


  if (argc < 3)
    Help (progname);

  if (getenv ("PILOTPORT"))
    {
      strcpy (addr.pi_device, getenv ("PILOTPORT"));
    }
  else
    {
      strcpy (addr.pi_device, PILOTPORT);
    }

  while (((c = getopt (argc, argv, "DTeqp:t:d:c:arw")) != -1) && (mode == 0))
    {
      switch (c)
        {

        case 't':
          tableformat = 1;
          tabledelim = atoi (optarg);
          break;
        case 'D':
          deleteallcategories = 1;
          break;
        case 'T':
          tablehead = 1;
          break;
        case 'a':
          augment = 1;
          break;
        case 'q':
          quiet = 1;
          break;
        case 'p':
          /* optarg is name of port to use instead of
             $PILOTPORT or /dev/pilot */

          strcpy (addr.pi_device, optarg);
          break;
        case 'd':
          deletecategory = optarg;
          break;
        case 'e':
          encodechars = 1;
          break;
        case 'c':
          defaultcategoryname = optarg;
          break;
        case 'r':
          mode = 1;
          break;
        case 'w':
          mode = 2;
          break;
        case 'h':
        case '?':
          Help (progname);
        }
    }

  if (mode == 0)
    Help (progname);

#if 0
  if (!(sd = pi_socket (PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP)))
    {
      perror ("pi_socket");
      exit (1);
    }

  addr.pi_family = PI_AF_SLP;
/*   strcpy(addr.pi_device, device); */

  ret = pi_bind (sd, (struct sockaddr *) &addr, sizeof (addr));
  if (ret == -1)
    {
      fprintf (stderr, "\n   Unable to bind to port %s\n", addr.pi_device);
      perror ("   pi_bind");
      fprintf (stderr, "\n");
      exit (1);
    }

  if (!quiet)
    {
      printf
        ("   Port: %s\n\n   Please press the HotSync button now...\n",
         addr.pi_device);
    }

  ret = pi_listen (sd, 1);
  if (ret == -1)
    {
      fprintf (stderr, "\n   Error listening on %s\n", addr.pi_device);
      perror ("   pi_listen");
      fprintf (stderr, "\n");
      exit (1);
    }

  sd = pi_accept (sd, 0, 0);
  if (sd == -1)
    {
      fprintf (stderr, "\n   Error accepting data on %s\n", addr.pi_device);
      perror ("   pi_accept");
      fprintf (stderr, "\n");
      exit (1);
    }

  if (!quiet)
    {
      fprintf (stderr, "Connected...\n");
    }
#endif


  /* Ask the pilot who it is. */
  dlp_ReadUserInfo (sd, &U);

  /* Tell user (via the Palm) that we are starting things up */
  dlp_OpenConduit (sd);

  /* Open the MemoDB.pdb database, store access handle in db */
  if (dlp_OpenDB (sd, 0, 0x80 | 0x40, "AddressDB", &db) < 0)
    {
      puts ("Unable to open AddressDB");
      dlp_AddSyncLogEntry (sd, "Unable to open AddressDB.\n");
      exit (1);
    }

  l = dlp_ReadAppBlock (sd, db, 0, (unsigned char *) buf, 0xffff);
  unpack_AddressAppInfo (&aai, (unsigned char *) buf, l);

  if (defaultcategoryname)
    defaultcategory = match_category (defaultcategoryname, &aai);
  else
    defaultcategory = 0;        /* Unfiled */

  if (mode == 2)
    {                           /* Write */
#if 0
      FILE *f = fopen (argv[optind], "w");

      if (f == NULL)
        {
          sprintf (buf, "%s: %s", argv[0], argv[optind]);
          perror (buf);
          exit (1);
        }
#endif
      write_file (NULL, sd, db, &aai);
      if (deletecategory)
        dlp_DeleteCategory (sd, db, match_category (deletecategory, &aai));
#if 0
      fclose (f);
#endif
    }
  else if (mode == 1)
    {
      FILE *f;

      while (optind < argc)
        {
#if 0
          f = fopen (argv[optind], "r");
          if (f == NULL)
            {
              sprintf (buf, "%s: %s", argv[0], argv[optind]);
              perror (buf);
              continue;
            }
#endif
          if (deletecategory)
            dlp_DeleteCategory (sd, db,
                                match_category (deletecategory, &aai));

          if (deleteallcategories)
            {
              int i;

              for (i = 0; i < 16; i++)
                if (strlen (aai.category.name[i]) > 0)
                  dlp_DeleteCategory (sd, db, i);
            }

          emdb_dump (sd, db, &aai);
#if 0
          read_file (NULL, sd, db, &aai);
          fclose (f);
#endif
          optind++;
        }
    }

  /* Close the database */
  dlp_CloseDB (sd, db);

  /* Tell the user who it is, with a different PC id. */
#if  0
  U.lastSyncPC = 0xDEADBEEF;
  U.successfulSyncDate = time (NULL);
  U.lastSyncDate = U.successfulSyncDate;
  dlp_WriteUserInfo (sd, &U);
#endif

  if (mode == 1)
    {
      dlp_AddSyncLogEntry (sd, "Wrote addresses to Palm.\n");
    }
  else if (mode == 2)
    {
      dlp_AddSyncLogEntry (sd, "Read addresses from Palm.\n");
    }

  /* All of the following code is now unnecessary, but harmless */

#if 0
  dlp_EndOfSync (sd, 0);
  pi_close (sd);
#endif

  return 0;
}
